#!/usr/bin/env python3

# ruff: noqa: I001

from openg2p_spar_mapper_api.app import Initializer

from openg2p_fastapi_common.ping import PingInitializer


initializer = Initializer()
PingInitializer()

app = initializer.return_app()

if __name__ == "__main__":
    initializer.main()



from openg2p_fastapi_common.service import BaseService
from openg2p_g2pconnect_common_lib.schemas import StatusEnum
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkStatusReasonCode,
    ResolveStatusReasonCode,
    SingleLinkRequest,
    SingleResolveRequest,
    SingleUnlinkRequest,
    SingleUpdateRequest,
    UnlinkStatusReasonCode,
    UpdateStatusReasonCode,
)
from sqlalchemy import and_, select

from ..models import IdFaMapping
from .exceptions import (
    LinkValidationException,
    ResolveValidationException,
    UnlinkValidationException,
    UpdateValidationException,
)


class IdFaMappingValidations(BaseService):
    async def validate_link_request(
        self, connection, single_link_request: SingleLinkRequest
    ) -> None:
        # Check if the ID is null
        if not single_link_request.id:
            raise LinkValidationException(
                message="ID is null",
                status=StatusEnum.rjct,
                validation_error_type=LinkStatusReasonCode.rjct_id_invalid,
            )

        # Check if the FA is null
        if not single_link_request.fa:
            raise LinkValidationException(
                message="FA is null",
                status=StatusEnum.rjct,
                validation_error_type=LinkStatusReasonCode.rjct_fa_invalid,
            )

        # Check if the ID is already mapped
        result = await connection.execute(
            select(IdFaMapping).where(
                IdFaMapping.id_value == single_link_request.id,
            )
        )
        link_request_from_db = result.first()

        if link_request_from_db:
            raise LinkValidationException(
                message="ID and FA are already mapped",
                status=StatusEnum.rjct,
                validation_error_type=LinkStatusReasonCode.rjct_reference_id_duplicate,
            )

        return None

    async def validate_update_request(
        self, connection, single_update_request: SingleUpdateRequest
    ) -> None:
        if not single_update_request.id:
            raise UpdateValidationException(
                message="ID is null",
                status=StatusEnum.rjct,
                validation_error_type=UpdateStatusReasonCode.rjct_id_invalid,
            )

        if not single_update_request.fa:
            raise UpdateValidationException(
                message="FA is null",
                status=StatusEnum.rjct,
                validation_error_type=UpdateStatusReasonCode.rjct_fa_invalid,
            )

        result = await connection.execute(
            select(IdFaMapping).where(
                and_(
                    IdFaMapping.id_value == single_update_request.id,
                )
            )
        )
        update_request_from_db = result.first()

        if not update_request_from_db:
            raise UpdateValidationException(
                message="ID doesnt exist please link first",
                status=StatusEnum.rjct,
                validation_error_type=UpdateStatusReasonCode.rjct_reference_id_duplicate,
            )

        return None

    async def validate_resolve_request(
        self, single_resolve_request: SingleResolveRequest
    ) -> None:
        if not single_resolve_request.id and not single_resolve_request.fa:
            raise ResolveValidationException(
                message="ID is required",
                status=StatusEnum.rjct,
                validation_error_type=ResolveStatusReasonCode.rjct_reference_id_invalid,
            )
        return None

    async def validate_unlink_request(
        self, connection, single_unlink_request: SingleUnlinkRequest
    ) -> None:
        if not single_unlink_request.id:
            raise UnlinkValidationException(
                message="ID is null",
                status=StatusEnum.rjct,
                validation_error_type=UnlinkStatusReasonCode.rjct_id_invalid,
            )

        result = await connection.execute(
            select(IdFaMapping).where(
                IdFaMapping.id_value == single_unlink_request.id,
            )
        )

        if single_unlink_request.fa:
            result = await connection.execute(
                select(IdFaMapping).where(
                    IdFaMapping.id_value == single_unlink_request.id,
                    IdFaMapping.fa_value == single_unlink_request.fa,
                )
            )

        unlink_request_from_db = result.first()

        if not unlink_request_from_db:
            raise UnlinkValidationException(
                message="ID doesnt exist please link first",
                status=StatusEnum.rjct,
                validation_error_type=UnlinkStatusReasonCode.rjct_id_invalid,
            )

        return None



from datetime import datetime

from openg2p_fastapi_common.service import BaseService
from openg2p_g2pconnect_common_lib.schemas import (
    Request,
)
from openg2p_g2pconnect_mapper_lib.schemas.link import (
    LinkRequest,
)

from ..models import IdFaMapping


class SyncRequestHelper(BaseService):
    @staticmethod
    def deconstruct_link_request(self, request: Request) -> list[IdFaMapping]:
        linkRequest: LinkRequest = LinkRequest.model_validate(request.message)
        return [
            IdFaMapping(
                id_value=link_req.id,
                fa_value=link_req.fa,
                name=link_req.name,
                phone=link_req.phone_number,
                additional_info=link_req.additional_info,
                active=True,
                created_at=datetime.now(),
            )
            for link_req in linkRequest.link_request
        ]


class AsyncRequestHelper(BaseService):
    def deconstruct_link_request(self, request: Request) -> list[IdFaMapping]:
        linkRequest: LinkRequest = LinkRequest.model_validate(request.message)
        return [
            IdFaMapping(
                id_value=link_req.id,
                fa_value=link_req.fa,
                name=link_req.name,
                phone=link_req.phone_number,
                additional_info=link_req.additional_info,
                active=True,
                created_at=datetime.now(),
            )
            for link_req in linkRequest.link_request
        ]



from typing import Any, Dict, List, Optional

from openg2p_fastapi_common.context import dbengine
from openg2p_fastapi_common.models import BaseORMModelWithTimes
from sqlalchemy import JSON, String, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, mapped_column


class IdFaMapping(BaseORMModelWithTimes):
    __tablename__ = "id_fa_mappings"

    id_value: Mapped[str] = mapped_column(String(), index=True, unique=True)
    fa_value: Mapped[str] = mapped_column(String(), index=True)

    name: Mapped[Optional[str]] = mapped_column(String())
    phone: Mapped[Optional[str]] = mapped_column(String())

    additional_info: Mapped[Optional[List[Dict[str, Any]]]] = mapped_column(
        JSON(), default=None
    )

    @classmethod
    async def get_all_by_query(cls, session=None, **kwargs):
        if "active" not in kwargs:
            kwargs["active"] = True
        response = []

        original_session = False
        if not session:
            original_session = True
            session = AsyncSession(dbengine.get())

        stmt = select(cls)
        for key, value in kwargs.items():
            if value is not None:
                stmt = stmt.where(getattr(cls, key) == value)

        stmt = stmt.order_by(cls.id.asc())

        result = await session.execute(stmt)

        response = list(result.scalars())

        if not original_session:
            await session.aclose()

        return response



from typing import Optional

from openg2p_g2pconnect_common_lib.config import Settings as BaseSettings
from pydantic import AnyUrl
from pydantic_settings import SettingsConfigDict

from . import __version__


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="spar_mapper_", env_file=".env", extra="allow"
    )

    openapi_title: str = "OpenG2P SPAR Account Mapper"
    openapi_description: str = """
    This module maps the beneficiary ID to a Financial Address.

    ***********************************
    Further details goes here
    ***********************************
    """
    openapi_version: str = __version__

    db_dbname: str = "openg2p_spar_db"

    default_callback_url: Optional[AnyUrl] = None
    default_callback_timeout: int = 10
    callback_sender_id: str = "mapper.dev.openg2p.net"



from datetime import datetime

from openg2p_fastapi_common.service import BaseService
from openg2p_g2pconnect_common_lib.schemas import (
    AsyncAck,
    AsyncCallbackRequest,
    AsyncCallbackRequestHeader,
    AsyncResponse,
    AsyncResponseMessage,
    Request,
    StatusEnum,
    SyncResponse,
    SyncResponseHeader,
)
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    LinkRequestMessage,
    LinkResponse,
    LinkResponseMessage,
    ResolveRequest,
    ResolveRequestMessage,
    ResolveResponse,
    ResolveResponseMessage,
    SingleLinkResponse,
    SingleResolveResponse,
    SingleUnlinkResponse,
    SingleUpdateResponse,
    UnlinkRequest,
    UnlinkRequestMessage,
    UnlinkResponse,
    UnlinkResponseMessage,
    UpdateRequest,
    UpdateRequestMessage,
    UpdateResponse,
    UpdateResponseMessage,
)

from .exceptions import (
    RequestValidationException,
)


class SyncResponseHelper(BaseService):
    def construct_success_sync_link_response(
        self,
        link_request: LinkRequest,
        single_link_responses: list[SingleLinkResponse],
    ) -> LinkResponse:
        link_request_message: LinkRequestMessage = link_request.message
        link_response_message: LinkResponseMessage = LinkResponseMessage(
            transaction_id=link_request_message.transaction_id,
            correlation_id=None,
            link_response=single_link_responses,
        )
        total_count = len(link_response_message.link_response)
        completed_count = len(
            [
                link
                for link in link_response_message.link_response
                if link.status == StatusEnum.succ
            ]
        )
        return LinkResponse(
            header=SyncResponseHeader(
                version="1.0.0",
                message_id=link_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=link_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=link_request.header.sender_id,
                receiver_id=link_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=link_response_message,
        )

    def construct_success_sync_update_response(
        self,
        update_request: UpdateRequest,
        single_update_responses: list[SingleUpdateResponse],
    ) -> UpdateResponse:
        update_request_message: UpdateRequestMessage = update_request.message
        update_response_message: UpdateResponseMessage = UpdateResponseMessage(
            transaction_id=update_request_message.transaction_id,
            correlation_id=None,
            update_response=single_update_responses,
        )
        total_count = len(update_response_message.update_response)
        completed_count = len(
            [
                update
                for update in update_response_message.update_response
                if update.status == StatusEnum.succ
            ]
        )
        return UpdateResponse(
            header=SyncResponseHeader(
                version="1.0.0",
                message_id=update_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=update_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=update_request.header.sender_id,
                receiver_id=update_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=update_response_message,
        )

    def construct_success_sync_resolve_response(
        self,
        resolve_request: ResolveRequest,
        single_resolve_responses: list[SingleResolveResponse],
    ) -> ResolveResponse:
        resolve_request_message: ResolveRequestMessage = resolve_request.message
        resolve_response_message: ResolveResponseMessage = ResolveResponseMessage(
            transaction_id=resolve_request_message.transaction_id,
            correlation_id=None,
            resolve_response=single_resolve_responses,
        )
        total_count = len(resolve_response_message.resolve_response)
        completed_count = len(
            [
                resolve
                for resolve in resolve_response_message.resolve_response
                if resolve.status == StatusEnum.succ
            ]
        )
        return ResolveResponse(
            header=SyncResponseHeader(
                version="1.0.0",
                message_id=resolve_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=resolve_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=resolve_request.header.sender_id,
                receiver_id=resolve_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=resolve_response_message,
        )

    def construct_success_sync_unlink_response(
        self,
        unlink_request: UnlinkRequest,
        single_unlink_responses: list[SingleUnlinkResponse],
    ) -> UnlinkResponse:
        unlink_request_message: UnlinkRequestMessage = unlink_request.message
        unlink_response_message: UnlinkResponseMessage = UnlinkResponseMessage(
            transaction_id=unlink_request_message.transaction_id,
            correlation_id=None,
            unlink_response=single_unlink_responses,
        )
        total_count = len(unlink_response_message.unlink_response)
        completed_count = len(
            [
                unlink
                for unlink in unlink_response_message.unlink_response
                if unlink.status == StatusEnum.succ
            ]
        )
        return UnlinkResponse(
            header=SyncResponseHeader(
                version="1.0.0",
                message_id=unlink_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=unlink_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=unlink_request.header.sender_id,
                receiver_id=unlink_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=unlink_response_message,
        )

    def construct_error_sync_response(
        self, request: Request, exception: RequestValidationException
    ) -> SyncResponse:
        return SyncResponse(
            signature=None,
            header=SyncResponseHeader(
                version="1.0.0",
                message_id=request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=request.header.action,
                status=StatusEnum.rjct,
                status_reason_code=exception.code,
                status_reason_message=exception.message,
            ),
            message={},
        )


class AsyncResponseHelper(BaseService):
    def construct_success_async_response(
        self,
        request: Request,
        correlation_id: str,
    ) -> AsyncResponse:
        return AsyncResponse(
            message=AsyncResponseMessage(
                ack_status=AsyncAck.ACK,
                correlation_id=correlation_id,
                timestamp=datetime.utcnow(),
            )
        )

    def construct_error_async_response(
        self,
        request: Request,
        exception: RequestValidationException,
    ) -> AsyncResponse:
        return AsyncResponse(
            message=AsyncResponseMessage(
                ack_status=AsyncAck.NACK,
                timestamp=datetime.utcnow(),
                error={
                    "code": exception.code,
                    "message": exception.message,
                },
            )
        )

    def construct_success_async_callback_link_request(
        self,
        link_request: LinkRequest,
        correlation_id: str,
        single_link_responses: list[SingleLinkResponse],
    ) -> AsyncCallbackRequest:
        total_count = len(single_link_responses)
        completed_count = len(
            [link for link in single_link_responses if link.status == StatusEnum.succ]
        )
        link_request_message: LinkRequestMessage = link_request.message

        link_response_message: LinkResponseMessage = LinkResponseMessage(
            transaction_id=link_request_message.transaction_id,
            correlation_id=None,
            link_response=single_link_responses,
        )
        return AsyncCallbackRequest(
            signature=None,
            header=AsyncCallbackRequestHeader(
                version="1.0.0",
                message_id=link_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=link_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=link_request.header.sender_id,
                receiver_id=link_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=link_response_message,
        )

    def construct_success_async_callback_update_request(
        self,
        update_request: UpdateRequest,
        correlation_id: str,
        single_update_responses: list[SingleUpdateResponse],
    ) -> AsyncCallbackRequest:
        total_count = len(single_update_responses)
        completed_count = len(
            [
                update
                for update in single_update_responses
                if update.status == StatusEnum.succ
            ]
        )
        update_request_message: UpdateRequestMessage = update_request.message
        update_response_message: UpdateResponseMessage = UpdateResponseMessage(
            transaction_id=update_request_message.transaction_id,
            correlation_id=None,
            update_response=single_update_responses,
        )
        return AsyncCallbackRequest(
            signature=None,
            header=AsyncCallbackRequestHeader(
                version="1.0.0",
                message_id=update_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=update_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=update_request.header.sender_id,
                receiver_id=update_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=update_response_message,
        )

    def construct_success_async_callback_resolve_request(
        self,
        resolve_request: ResolveRequest,
        correlation_id: str,
        single_resolve_responses: list[SingleResolveResponse],
    ) -> AsyncCallbackRequest:
        total_count = len(single_resolve_responses)
        completed_count = len(
            [
                resolve
                for resolve in single_resolve_responses
                if resolve.status == StatusEnum.succ
            ]
        )
        resolve_request_message: ResolveRequestMessage = resolve_request.message
        resolve_response_message: ResolveResponseMessage = ResolveResponseMessage(
            transaction_id=resolve_request_message.transaction_id,
            correlation_id=None,
            resolve_response=single_resolve_responses,
        )
        return AsyncCallbackRequest(
            signature=None,
            header=AsyncCallbackRequestHeader(
                version="1.0.0",
                message_id=resolve_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=resolve_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=resolve_request.header.sender_id,
                receiver_id=resolve_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=resolve_response_message,
        )

    def construct_success_async_callback_unlink_request(
        self,
        unlink_request: UnlinkRequest,
        correlation_id: str,
        single_unlink_responses: list[SingleUnlinkResponse],
    ) -> AsyncCallbackRequest:
        total_count = len(single_unlink_responses)
        completed_count = len(
            [
                unlink
                for unlink in single_unlink_responses
                if unlink.status == StatusEnum.succ
            ]
        )
        unlink_request_message: UnlinkRequestMessage = unlink_request.message
        unlink_response_message: UnlinkResponseMessage = UnlinkResponseMessage(
            transaction_id=unlink_request_message.transaction_id,
            correlation_id=None,
            unlink_response=single_unlink_responses,
        )
        return AsyncCallbackRequest(
            signature=None,
            header=AsyncCallbackRequestHeader(
                version="1.0.0",
                message_id=unlink_request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=unlink_request.header.action,
                status=StatusEnum.succ,
                status_reason_code=None,
                status_reason_message=None,
                total_count=total_count,
                completed_count=completed_count,
                sender_id=unlink_request.header.sender_id,
                receiver_id=unlink_request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message=unlink_response_message,
        )

    def construct_error_async_callback_request(
        self,
        request: Request,
        exception: RequestValidationException,
    ) -> AsyncCallbackRequest:
        return AsyncCallbackRequest(
            signature=None,
            header=AsyncCallbackRequestHeader(
                version="1.0.0",
                message_id=request.header.message_id,
                message_ts=datetime.now().isoformat(),
                action=request.header.action,
                status=StatusEnum.rjct,
                status_reason_code=exception.code,
                status_reason_message=exception.message,
                total_count=0,
                completed_count=0,
                sender_id=request.header.sender_id,
                receiver_id=request.header.receiver_id,
                is_msg_encrypted=False,
                meta={},
            ),
            message={},
        )



import logging
from datetime import datetime

from openg2p_fastapi_common.context import dbengine
from openg2p_fastapi_common.service import BaseService
from openg2p_g2pconnect_common_lib.schemas import StatusEnum
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    LinkRequestMessage,
    LinkStatusReasonCode,
    ResolveRequest,
    ResolveRequestMessage,
    ResolveScope,
    ResolveStatusReasonCode,
    SingleLinkResponse,
    SingleResolveRequest,
    SingleResolveResponse,
    SingleUnlinkResponse,
    SingleUpdateRequest,
    SingleUpdateResponse,
    UnlinkRequest,
    UnlinkRequestMessage,
    UpdateRequest,
    UpdateRequestMessage,
    UpdateStatusReasonCode,
)
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker

from ..config import Settings
from ..models import IdFaMapping
from ..services.exceptions import (
    LinkValidationException,
    ResolveValidationException,
    UnlinkValidationException,
    UpdateValidationException,
)
from ..services.id_fa_mapping_validations import IdFaMappingValidations
from ..services.session_service import SessionInitializer

_config = Settings.get_config()
_logger = logging.getLogger(_config.logging_default_logger_name)


class MapperService(BaseService):
    async def link(self, link_request: LinkRequest):
        session_maker = async_sessionmaker(dbengine.get(), expire_on_commit=False)
        async with session_maker() as session:
            link_request_message: LinkRequestMessage = link_request.message
            mappings_to_add = []
            single_link_responses: list[SingleLinkResponse] = []

            for single_link_request in link_request_message.link_request:
                try:
                    await IdFaMappingValidations.get_component().validate_link_request(
                        connection=session, single_link_request=single_link_request
                    )

                    mappings_to_add.append(
                        self.construct_id_fa_mapping(single_link_request)
                    )
                    single_link_responses.append(
                        self.construct_single_link_response_for_success(
                            single_link_request
                        )
                    )
                except LinkValidationException as e:
                    LinkValidationException(
                        message="Duplicate ID exists. Use 'update' instead.",
                        status=StatusEnum.rjct,
                        validation_error_type=LinkStatusReasonCode.rjct_id_invalid,
                    )
                    single_link_responses.append(
                        self.construct_single_link_response_for_failure(
                            single_link_request, e
                        )
                    )
        session.add_all(mappings_to_add)
        await session.commit()
        return single_link_responses

    def construct_id_fa_mapping(self, single_link_request):
        return IdFaMapping(
            id_value=single_link_request.id,
            fa_value=single_link_request.fa,
            name=single_link_request.name,
            phone=single_link_request.phone_number,
            additional_info=single_link_request.additional_info,
            active=True,
        )

    def construct_single_link_response_for_success(self, single_link_request):
        return SingleLinkResponse(
            reference_id=single_link_request.reference_id,
            timestamp=datetime.now(),
            fa=single_link_request.fa,
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message=None,
            additional_info=None,
            locale=single_link_request.locale,
        )

    def construct_single_link_response_for_failure(self, single_link_request, error):
        return SingleLinkResponse(
            reference_id=single_link_request.reference_id,
            timestamp=datetime.now(),
            fa=single_link_request.fa,
            status=StatusEnum.rjct,
            status_reason_code=error.validation_error_type,
            status_reason_message=error.message,
            additional_info=None,
            locale=single_link_request.locale,
        )

    async def update(self, update_request: UpdateRequest):
        session_maker = async_sessionmaker(dbengine.get(), expire_on_commit=False)
        async with session_maker() as session:
            update_request_message: UpdateRequestMessage = update_request.message
            single_update_responses: list[SingleUpdateResponse] = []

            for single_update_request in update_request_message.update_request:
                try:
                    await IdFaMappingValidations.get_component().validate_update_request(
                        connection=session, single_update_request=single_update_request
                    )

                    single_update_responses.append(
                        self.construct_single_update_response_for_success(
                            single_update_request
                        )
                    )

                    await self.update_mapping(session, single_update_request)

                except UpdateValidationException as e:
                    UpdateValidationException(
                        message="Mapping doesn't exist against the given ID.",
                        status=StatusEnum.rjct,
                        validation_error_type=UpdateStatusReasonCode.rjct_id_invalid,
                    )
                    single_update_responses.append(
                        self.construct_single_update_response_for_failure(
                            single_update_request, e
                        )
                    )

        await session.commit()

        return single_update_responses

    async def update_mapping(self, session, single_update_request):
        single_update_request: SingleUpdateRequest = SingleUpdateRequest.model_validate(
            single_update_request
        )
        single_response = self.construct_single_update_response_for_success(
            single_update_request
        )
        result = await session.execute(
            select(IdFaMapping).where(IdFaMapping.id_value == single_update_request.id)
        )
        result = result.scalar()

        if result:
            if single_update_request.fa:
                result.fa_value = single_update_request.fa
            if single_update_request.name:
                result.name = single_update_request.name
            if single_update_request.phone_number:
                result.phone = single_update_request.phone_number
            if single_update_request.additional_info:
                result.additional_info = single_update_request.additional_info
        else:
            single_response.status = StatusEnum.rjct
            single_response.status_reason_code = UpdateStatusReasonCode.rjct_id_invalid
            single_response.status_reason_message = (
                "Mapping doesnt exist against given ID. Use 'link' instead."
            )
        await session.commit()

    def construct_single_update_response_for_success(self, single_update_request):
        return SingleUpdateResponse(
            id=single_update_request.id,
            reference_id=single_update_request.reference_id,
            timestamp=datetime.now(),
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message=None,
            additional_info=None,
            locale=single_update_request.locale,
        )

    def construct_single_update_response_for_failure(
        self, single_update_request, error
    ):
        return SingleUpdateResponse(
            reference_id=single_update_request.reference_id,
            timestamp=datetime.now(),
            fa=single_update_request.fa,
            status=StatusEnum.rjct,
            status_reason_code=error.validation_error_type,
            status_reason_message=error.message,
            additional_info=None,
            locale=single_update_request.locale,
        )

    async def resolve(self, resolve_request: ResolveRequest):
        session_initializer = SessionInitializer.get_component()
        session: AsyncSession = await session_initializer.get_session_from_pool()
        async with session.begin():
            resolve_request_message: ResolveRequestMessage = resolve_request.message

            # Collect all ID values for bulk query
            id_values = [
                single_resolve_request.id
                for single_resolve_request in resolve_request_message.resolve_request
            ]

            # Validate all requests and collect validated requests
            validated_requests = []
            single_resolve_responses = []
            for single_resolve_request in resolve_request_message.resolve_request:
                try:
                    await IdFaMappingValidations.get_component().validate_resolve_request(
                        single_resolve_request=single_resolve_request,
                    )
                    validated_request = SingleResolveRequest.model_validate(
                        single_resolve_request
                    )
                    validated_requests.append(validated_request)
                except ResolveValidationException as e:
                    single_resolve_responses.append(
                        self.construct_single_resolve_response_for_failure(
                            single_resolve_request, e
                        )
                    )

            # Construct and execute bulk query
            if validated_requests:
                stmt, results = await self.construct_bulk_query(session, id_values)

                # Create a dictionary for fast lookup of results by ID
                result_dict = {result.id_value: result for result in results}

                # Create responses for all validated requests
                for validated_request in validated_requests:
                    result = result_dict.get(validated_request.id)
                    if result:
                        single_resolve_response = self.construct_single_resolve(
                            validated_request, result
                        )
                        single_resolve_responses.append(
                            self.construct_single_resolve_response_for_success(
                                single_resolve_response
                            )
                        )
                    else:
                        resolve_validation_exception = ResolveValidationException(
                            message="ID doesn't exist, please link first",
                            status=StatusEnum.succ,
                            validation_error_type=ResolveStatusReasonCode.succ_fa_not_linked_to_id,
                        )
                        single_resolve_responses.append(
                            self.construct_single_resolve_response_for_failure(
                                validated_request, resolve_validation_exception
                            )
                        )

            await session.commit()
        return single_resolve_responses

    def construct_single_resolve(
        self, single_resolve_request, result
    ) -> SingleResolveResponse:
        single_response = self.construct_single_resolve_response_for_success(
            single_resolve_request
        )
        if result:
            if single_resolve_request.scope == ResolveScope.details:
                single_response.status = StatusEnum.succ
                single_response.additional_info = result.additional_info
                single_response.fa = result.fa_value
                single_response.id = result.id_value
                single_response.status_reason_code = (
                    ResolveStatusReasonCode.succ_id_active
                )
            else:
                single_response.status = StatusEnum.succ
                single_response.status_reason_code = (
                    ResolveStatusReasonCode.succ_id_active
                )
        else:
            single_response.status = StatusEnum.succ
            single_response.status_reason_code = (
                ResolveStatusReasonCode.succ_id_not_found
            )
            single_response.status_reason_message = (
                "Mapping not found against given ID."
            )
        return single_response

    async def construct_query(self, session, single_resolve_request):
        self.construct_single_resolve_response_for_success(single_resolve_request)
        stmt = None
        id_query = IdFaMapping.id_value == single_resolve_request.id
        stmt = select(IdFaMapping).where(id_query)
        result = await session.execute(stmt)
        result = result.scalar()
        return stmt, result

    async def construct_bulk_query(self, session, id_values):
        stmt = None
        id_query = IdFaMapping.id_value.in_(id_values)
        stmt = select(IdFaMapping).where(id_query)
        result = await session.execute(stmt)
        result = result.scalars().all()
        return stmt, result

    def construct_single_resolve_response_for_success(self, single_resolve_request):
        return SingleResolveResponse(
            id=single_resolve_request.id,
            reference_id=single_resolve_request.reference_id,
            timestamp=datetime.now(),
            fa=single_resolve_request.fa,
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message=None,
            additional_info=single_resolve_request.additional_info,
            locale=single_resolve_request.locale,
        )

    def construct_single_resolve_response_for_failure(
        self, single_resolve_request, error
    ):
        return SingleResolveResponse(
            reference_id=single_resolve_request.reference_id,
            timestamp=datetime.now(),
            fa=single_resolve_request.fa,
            status=error.status,
            status_reason_code=error.validation_error_type,
            status_reason_message=error.message,
            additional_info=None,
            locale=single_resolve_request.locale,
        )

    async def unlink(self, unlink_request: UnlinkRequest):
        session_maker = async_sessionmaker(dbengine.get(), expire_on_commit=False)
        async with session_maker() as session:
            unlink_request_message: UnlinkRequestMessage = unlink_request.message
            single_unlink_responses: list[SingleUnlinkResponse] = []
            for single_unlink_request in unlink_request_message.unlink_request:
                try:
                    await IdFaMappingValidations.get_component().validate_unlink_request(
                        connection=session, single_unlink_request=single_unlink_request
                    )
                    await session.execute(
                        delete(IdFaMapping).where(
                            IdFaMapping.id_value == single_unlink_request.id
                        )
                    )
                    single_unlink_responses.append(
                        self.construct_single_unlink_response_for_success(
                            single_unlink_request
                        )
                    )
                except UnlinkValidationException as e:
                    single_unlink_responses.append(
                        self.construct_single_unlink_response_for_failure(
                            single_unlink_request, e
                        )
                    )
            await session.commit()
        return single_unlink_responses

    def construct_single_unlink_response_for_success(self, single_unlink_request):
        return SingleUnlinkResponse(
            reference_id=single_unlink_request.reference_id,
            timestamp=datetime.now(),
            fa=single_unlink_request.fa,
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message=None,
            additional_info=None,
            locale=single_unlink_request.locale,
        )

    def construct_single_unlink_response_for_failure(
        self, single_unlink_request, error
    ):
        return SingleUnlinkResponse(
            reference_id=single_unlink_request.reference_id,
            timestamp=datetime.now(),
            fa=single_unlink_request.fa,
            status=StatusEnum.rjct,
            status_reason_code=error.validation_error_type,
            status_reason_message=error.message,
            additional_info=None,
            locale=single_unlink_request.locale,
        )



from openg2p_fastapi_common.context import dbengine
from openg2p_fastapi_common.service import BaseService
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker


class SessionInitializer(BaseService):
    def __init__(self):
        super().__init__("SessionInitializer")
        self.session_maker = None

    async def get_session_from_pool(self) -> AsyncSession:
        if not self.session_maker:
            self.session_maker = async_sessionmaker(
                dbengine.get(), expire_on_commit=False
            )

        async with self.session_maker() as session:
            return session



# ruff: noqa: E402
import asyncio

from .config import Settings

_config = Settings.get_config()

from openg2p_fastapi_common.app import Initializer as BaseInitializer
from openg2p_g2pconnect_common_lib.oauth_token import OAuthTokenService

from .controllers import (
    AsyncMapperController,
    SyncMapperController,
)
from .models import IdFaMapping
from .services import (
    AsyncRequestHelper,
    AsyncResponseHelper,
    IdFaMappingValidations,
    MapperService,
    RequestValidation,
    SessionInitializer,
    SyncRequestHelper,
    SyncResponseHelper,
)


class Initializer(BaseInitializer):
    def initialize(self, **kwargs):
        super().initialize()

        OAuthTokenService()
        SessionInitializer()
        MapperService()
        IdFaMappingValidations()
        SyncRequestHelper()
        AsyncRequestHelper()
        RequestValidation()
        SyncResponseHelper()
        AsyncResponseHelper()
        SyncMapperController().post_init()
        AsyncMapperController().post_init()

    def migrate_database(self, args):
        super().migrate_database(args)

        async def migrate():
            print("Migrating database")
            await IdFaMapping.create_migrate()

        asyncio.run(migrate())



import asyncio
from datetime import datetime
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from openg2p_g2pconnect_common_lib.schemas import (
    AsyncAck,
    AsyncCallbackRequest,
    AsyncCallbackRequestHeader,
    AsyncResponse,
    AsyncResponseMessage,
    RequestHeader,
    StatusEnum,
    SyncResponseStatusReasonCodeEnum,
)
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    LinkRequestMessage,
    SingleLinkRequest,
    SingleLinkResponse,
)
from openg2p_g2pconnect_mapper_lib.schemas.resolve import (
    ResolveRequest,
    ResolveRequestMessage,
    SingleResolveRequest,
    SingleResolveResponse,
)
from openg2p_g2pconnect_mapper_lib.schemas.unlink import (
    SingleUnlinkRequest,
    SingleUnlinkResponse,
    UnlinkRequest,
    UnlinkRequestMessage,
)
from openg2p_g2pconnect_mapper_lib.schemas.update import (
    SingleUpdateRequest,
    SingleUpdateResponse,
    UpdateRequest,
    UpdateRequestMessage,
)
from openg2p_spar_mapper_api.controllers.async_mapper_controller import (
    AsyncMapperController,
)
from openg2p_spar_mapper_api.services import (
    RequestValidationException,
)


# Setup side effect for validate_signature to raise RequestValidationException
def mock_validate_signature(is_signature_valid):
    if not is_signature_valid:
        raise RequestValidationException(
            code=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
            message=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
        )


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_link_async(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    # Setup MagicMock for MapperService and RequestValidation components
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.link = (
        AsyncMock()
    )  # link method should return an awaitable
    mock_request_validation_instance = MagicMock()

    # Assign return values to the get_component mocks
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )

    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    mock_async_response_helper_instance = MagicMock()
    expected_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="1234",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="ACK",
        )
    )
    mock_async_response_helper_instance.construct_success_async_response.return_value = (
        expected_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_link_request = LinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="link",
            sender_id="test_sender",
            total_count=1,
        ),
        message=LinkRequestMessage(
            transaction_id="test_transaction_id",
            link_request=[
                SingleLinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.link_async(
        mock_link_request, is_signature_valid=True
    )
    assert (
        actual_response == expected_response
    ), "The response did not match the expected response."
    assert actual_response.message.correlation_id == "1234"
    assert actual_response.message.ack_status == AsyncAck.ACK
    assert actual_response.message.timestamp == expected_response.message.timestamp


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_link_async_invalid_signature(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    # Setup MagicMock for MapperService and RequestValidation components
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.link = AsyncMock()
    mock_request_validation_instance = MagicMock()

    # Assign return values to the get_component mocks
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )

    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    # Setup MagicMock for AsyncResponseHelper component
    mock_async_response_helper_instance = MagicMock()
    error_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="error_correlation_id",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="NACK",
            error={
                "code": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
                "message": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
            },
        )
    )
    mock_async_response_helper_instance.construct_error_async_response.return_value = (
        error_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_link_request = LinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="link",
            sender_id="test_sender",
            total_count=1,
        ),
        message=LinkRequestMessage(
            transaction_id="test_transaction_id",
            link_request=[
                SingleLinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.link_async(
        mock_link_request, is_signature_valid=False
    )

    assert (
        actual_response == error_response
    ), "The response did not match the expected error response."
    assert actual_response.message.correlation_id == "error_correlation_id"
    assert actual_response.message.ack_status == AsyncAck.NACK
    assert actual_response.message.timestamp == error_response.message.timestamp
    assert (
        actual_response.message.error.code
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )
    assert (
        actual_response.message.error.message
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_handle_service_and_link_callback(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = AsyncMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    single_link_responses = [
        SingleLinkResponse(
            reference_id="test_ref",
            timestamp=datetime.utcnow(),
            status=StatusEnum.succ,
            locale="en",
        )
    ]
    mock_mapper_service_instance.link.return_value = single_link_responses

    mock_request_validation_instance = AsyncMock()
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_async_response_helper_instance = AsyncMock()
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    link_request = LinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="link",
            sender_id="test_sender",
            total_count=1,
        ),
        message=LinkRequestMessage(
            transaction_id="test_transaction_id",
            link_request=[
                SingleLinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    # Simulate the behavior without actual execution
    await controller.handle_service_and_link_callback(
        link_request, "correlation_id", "link"
    )

    mock_async_response_helper_instance.construct_success_async_callback_link_request.assert_called_once_with(
        link_request, "correlation_id", single_link_responses
    )

    callback_args = (
        mock_async_response_helper_instance.construct_success_async_callback_link_request.call_args
    )
    assert callback_args[0][0] == link_request
    assert callback_args[0][1] == "correlation_id"
    assert callback_args[0][2] == single_link_responses


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_update_async(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.update = AsyncMock()
    mock_request_validation_instance = MagicMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    mock_async_response_helper_instance = MagicMock()
    expected_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="1234",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="ACK",
        )
    )
    mock_async_response_helper_instance.construct_success_async_response.return_value = (
        expected_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_update_request = UpdateRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="update",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UpdateRequestMessage(
            transaction_id="test_transaction_id",
            update_request=[
                SingleUpdateRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.update_async(
        mock_update_request, is_signature_valid=True
    )
    assert (
        actual_response == expected_response
    ), "The response did not match the expected response."
    assert actual_response.message.correlation_id == "1234"
    assert actual_response.message.ack_status == AsyncAck.ACK
    assert actual_response.message.timestamp == expected_response.message.timestamp


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_update_async_invalid_signature(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    # Setup MagicMock for MapperService and RequestValidation components
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.update = AsyncMock()
    mock_request_validation_instance = MagicMock()

    # Assign return values to the get_component mocks
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )

    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    # Setup MagicMock for AsyncResponseHelper component
    mock_async_response_helper_instance = MagicMock()
    error_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="error_correlation_id",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="NACK",
            error={
                "code": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
                "message": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
            },
        )
    )
    mock_async_response_helper_instance.construct_error_async_response.return_value = (
        error_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_update_request = UpdateRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="update",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UpdateRequestMessage(
            transaction_id="test_transaction_id",
            update_request=[
                SingleUpdateRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.update_async(
        mock_update_request, is_signature_valid=False
    )

    assert (
        actual_response == error_response
    ), "The response did not match the expected error response."
    assert actual_response.message.correlation_id == "error_correlation_id"
    assert actual_response.message.ack_status == AsyncAck.NACK
    assert actual_response.message.timestamp == error_response.message.timestamp
    assert (
        actual_response.message.error.code
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )
    assert (
        actual_response.message.error.message
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_handle_service_and_update_callback(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = AsyncMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    single_update_responses = [
        SingleUpdateResponse(
            reference_id="test_ref",
            timestamp=datetime.utcnow(),
            status=StatusEnum.succ,
            locale="en",
        )
    ]
    mock_mapper_service_instance.update.return_value = single_update_responses

    mock_request_validation_instance = AsyncMock()
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_async_response_helper_instance = AsyncMock()
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    update_request = UpdateRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="update",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UpdateRequestMessage(
            transaction_id="test_transaction_id",
            update_request=[
                SingleUpdateRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )
    await controller.handle_service_and_update_callback(
        update_request, "correlation_id", "update"
    )

    mock_async_response_helper_instance.construct_success_async_callback_update_request.assert_called_once_with(
        update_request, "correlation_id", single_update_responses
    )

    callback_args = (
        mock_async_response_helper_instance.construct_success_async_callback_update_request.call_args
    )
    assert callback_args[0][0] == update_request
    assert callback_args[0][1] == "correlation_id"
    assert callback_args[0][2] == single_update_responses


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_resolve_async(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.resolve = AsyncMock()
    mock_request_validation_instance = MagicMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    mock_async_response_helper_instance = MagicMock()
    expected_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="1234",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="ACK",
        )
    )
    mock_async_response_helper_instance.construct_success_async_response.return_value = (
        expected_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_resolve_request = ResolveRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="resolve",
            sender_id="test_sender",
            total_count=1,
        ),
        message=ResolveRequestMessage(
            transaction_id="test_transaction_id",
            resolve_request=[
                SingleResolveRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.resolve_async(
        mock_resolve_request, is_signature_valid=True
    )
    assert (
        actual_response == expected_response
    ), "The response did not match the expected response."
    assert actual_response.message.correlation_id == "1234"
    assert actual_response.message.ack_status == AsyncAck.ACK
    assert actual_response.message.timestamp == expected_response.message.timestamp


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_resolve_async_invalid_signature(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    # Setup MagicMock for MapperService and RequestValidation components
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.resolve = AsyncMock()
    mock_request_validation_instance = MagicMock()

    # Assign return values to the get_component mocks
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )

    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    # Setup MagicMock for AsyncResponseHelper component
    mock_async_response_helper_instance = MagicMock()
    error_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="error_correlation_id",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="NACK",
            error={
                "code": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
                "message": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
            },
        )
    )
    mock_async_response_helper_instance.construct_error_async_response.return_value = (
        error_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_resolve_request = ResolveRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="resolve",
            sender_id="test_sender",
            total_count=1,
        ),
        message=ResolveRequestMessage(
            transaction_id="test_transaction_id",
            resolve_request=[
                SingleResolveRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.resolve_async(
        mock_resolve_request, is_signature_valid=False
    )

    assert (
        actual_response == error_response
    ), "The response did not match the expected error response."
    assert actual_response.message.correlation_id == "error_correlation_id"
    assert actual_response.message.ack_status == AsyncAck.NACK
    assert actual_response.message.timestamp == error_response.message.timestamp
    assert (
        actual_response.message.error.code
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )
    assert (
        actual_response.message.error.message
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_handle_service_and_resolve_callback(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = AsyncMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    single_resolve_responses = [
        SingleResolveResponse(
            reference_id="test_ref",
            timestamp=datetime.utcnow(),
            status=StatusEnum.succ,
            locale="en",
        )
    ]
    mock_mapper_service_instance.resolve.return_value = single_resolve_responses

    mock_request_validation_instance = AsyncMock()
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_async_response_helper_instance = AsyncMock()
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    resolve_request = ResolveRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="resolve",
            sender_id="test_sender",
            total_count=1,
        ),
        message=ResolveRequestMessage(
            transaction_id="test_transaction_id",
            resolve_request=[
                SingleResolveRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )
    await controller.handle_service_and_resolve_callback(
        resolve_request, "correlation_id", "resolve"
    )

    mock_async_response_helper_instance.construct_success_async_callback_resolve_request.assert_called_once_with(
        resolve_request, "correlation_id", single_resolve_responses
    )

    callback_args = (
        mock_async_response_helper_instance.construct_success_async_callback_resolve_request.call_args
    )
    assert callback_args[0][0] == resolve_request
    assert callback_args[0][1] == "correlation_id"
    assert callback_args[0][2] == single_resolve_responses


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_unlink_async(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.unlink = AsyncMock()
    mock_request_validation_instance = MagicMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    mock_async_response_helper_instance = MagicMock()
    expected_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="1234",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="ACK",
        )
    )
    mock_async_response_helper_instance.construct_success_async_response.return_value = (
        expected_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_unlink_request = UnlinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="unlink",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UnlinkRequestMessage(
            transaction_id="test_transaction_id",
            unlink_request=[
                SingleUnlinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.unlink_async(
        mock_unlink_request, is_signature_valid=True
    )
    assert (
        actual_response == expected_response
    ), "The response did not match the expected response."
    assert actual_response.message.correlation_id == "1234"
    assert actual_response.message.ack_status == AsyncAck.ACK
    assert actual_response.message.timestamp == expected_response.message.timestamp


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_unlink_async_invalid_signature(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    # Setup MagicMock for MapperService and RequestValidation components
    mock_mapper_service_instance = MagicMock()
    mock_mapper_service_instance.unlink = AsyncMock()
    mock_request_validation_instance = MagicMock()

    # Assign return values to the get_component mocks
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )

    mock_request_validation_instance.validate_signature.side_effect = (
        mock_validate_signature
    )

    # Setup MagicMock for AsyncResponseHelper component
    mock_async_response_helper_instance = MagicMock()
    error_response = AsyncResponse(
        message=AsyncResponseMessage(
            correlation_id="error_correlation_id",
            timestamp=datetime.utcnow().isoformat(),
            ack_status="NACK",
            error={
                "code": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
                "message": SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value,
            },
        )
    )
    mock_async_response_helper_instance.construct_error_async_response.return_value = (
        error_response
    )
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    mock_unlink_request = UnlinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="unlink",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UnlinkRequestMessage(
            transaction_id="test_transaction_id",
            unlink_request=[
                SingleUnlinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    actual_response = await controller.unlink_async(
        mock_unlink_request, is_signature_valid=False
    )

    assert (
        actual_response == error_response
    ), "The response did not match the expected error response."
    assert actual_response.message.correlation_id == "error_correlation_id"
    assert actual_response.message.ack_status == AsyncAck.NACK
    assert actual_response.message.timestamp == error_response.message.timestamp
    assert (
        actual_response.message.error.code
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )
    assert (
        actual_response.message.error.message
        == SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid.value
    )


@pytest.mark.asyncio
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.AsyncResponseHelper.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.RequestValidation.get_component"
)
@patch(
    "openg2p_spar_mapper_api.controllers.async_mapper_controller.MapperService.get_component"
)
async def test_handle_service_and_unlink_callback(
    mock_mapper_service_get_component,
    mock_request_validation_get_component,
    mock_async_response_helper_get_component,
):
    mock_mapper_service_instance = AsyncMock()
    mock_mapper_service_get_component.return_value = mock_mapper_service_instance
    single_unlink_responses = [
        SingleUnlinkResponse(
            reference_id="test_ref",
            timestamp=datetime.utcnow(),
            status=StatusEnum.succ,
            locale="en",
        )
    ]
    mock_mapper_service_instance.unlink.return_value = single_unlink_responses

    mock_request_validation_instance = AsyncMock()
    mock_request_validation_get_component.return_value = (
        mock_request_validation_instance
    )
    mock_async_response_helper_instance = AsyncMock()
    mock_async_response_helper_get_component.return_value = (
        mock_async_response_helper_instance
    )

    controller = AsyncMapperController()

    unlink_request = UnlinkRequest(
        header=RequestHeader(
            message_id="test_message_id",
            message_ts=datetime.utcnow().isoformat(),
            action="unlink",
            sender_id="test_sender",
            total_count=1,
        ),
        message=UnlinkRequestMessage(
            transaction_id="test_transaction_id",
            unlink_request=[
                SingleUnlinkRequest(
                    reference_id="test_ref",
                    timestamp=str(datetime.now()),
                    id="test_id",
                    fa="test_fa",
                )
            ],
        ),
    )

    await controller.handle_service_and_unlink_callback(
        unlink_request, "correlation_id", "unlink"
    )

    mock_async_response_helper_instance.construct_success_async_callback_unlink_request.assert_called_once_with(
        unlink_request, "correlation_id", single_unlink_responses
    )

    callback_args = (
        mock_async_response_helper_instance.construct_success_async_callback_unlink_request.call_args
    )
    assert callback_args[0][0] == unlink_request
    assert callback_args[0][1] == "correlation_id"
    assert callback_args[0][2] == single_unlink_responses


@pytest.mark.asyncio
async def test_make_callback():
    async_call_back_request = AsyncCallbackRequest(
        header=AsyncCallbackRequestHeader(
            message_id="123",
            message_ts="2021-05-01T12:00:00Z",
            action="test_action",
            status=StatusEnum.succ,
        ),
        message={"key": "value"},
    )

    url = "http://test.com/callback"
    url_suffix = "/suffix"

    with patch(
        "openg2p_spar_mapper_api.controllers.async_mapper_controller.httpx.post"
    ) as mock_post:
        mock_response = MagicMock()
        mock_response.raise_for_status.return_value = None
        mock_post.return_value = mock_response

        task = asyncio.ensure_future(
            AsyncMapperController.make_callback(
                async_call_back_request, url, url_suffix
            )
        )
        await task
        await asyncio.gather(
            *[t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        )

        mock_post.assert_called_once_with(
            f"{url.rstrip('/')}{url_suffix}",
            headers={"content-type": "application/json"},
            content=async_call_back_request.model_dump_json(),
            timeout=10,
        )



from typing import Annotated

from fastapi import Depends
from openg2p_fastapi_common.controller import BaseController
from openg2p_g2pconnect_common_lib.jwt_signature_validator import JWTSignatureValidator
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    LinkResponse,
    ResolveRequest,
    ResolveResponse,
    SingleLinkResponse,
    SingleResolveResponse,
    SingleUpdateResponse,
    UnlinkRequest,
    UnlinkResponse,
    UpdateRequest,
    UpdateResponse,
)

from ..services import (
    MapperService,
    RequestValidation,
    RequestValidationException,
    SyncResponseHelper,
)


class SyncMapperController(BaseController):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mapper_service = MapperService.get_component()

        self.router.prefix += "/sync"
        self.router.tags += ["G2PConnect Mapper Sync"]

        self.router.add_api_route(
            "/link",
            self.link_sync,
            responses={200: {"model": LinkResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/update",
            self.update_sync,
            responses={200: {"model": UpdateResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/resolve",
            self.resolve_sync,
            responses={200: {"model": ResolveResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/unlink",
            self.unlink_sync,
            responses={200: {"model": UnlinkResponse}},
            methods=["POST"],
        )

    async def link_sync(
        self,
        link_request: LinkRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
            RequestValidation.get_component().validate_request(link_request)
            RequestValidation.get_component().validate_link_request_header(link_request)
        except RequestValidationException as e:
            error_response = (
                SyncResponseHelper.get_component().construct_error_sync_response(
                    link_request, e
                )
            )
            return error_response

        single_link_responses: list[
            SingleLinkResponse
        ] = await self.mapper_service.link(link_request)
        return SyncResponseHelper.get_component().construct_success_sync_link_response(
            link_request,
            single_link_responses,
        )

    async def update_sync(
        self,
        update_request: UpdateRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
            RequestValidation.get_component().validate_request(update_request)
            RequestValidation.get_component().validate_update_request_header(
                update_request
            )
        except RequestValidationException as e:
            error_response = (
                SyncResponseHelper.get_component().construct_error_sync_response(
                    update_request, e
                )
            )
            return error_response

        single_update_responses: list[
            SingleUpdateResponse
        ] = await self.mapper_service.update(update_request)
        return (
            SyncResponseHelper.get_component().construct_success_sync_update_response(
                update_request,
                single_update_responses,
            )
        )

    async def resolve_sync(
        self,
        resolve_request: ResolveRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
            RequestValidation.get_component().validate_request(resolve_request)
            RequestValidation.get_component().validate_resolve_request_header(
                resolve_request
            )
        except RequestValidationException as e:
            error_response = (
                SyncResponseHelper.get_component().construct_error_sync_response(
                    resolve_request, e
                )
            )
            return error_response

        single_resolve_responses: list[
            SingleResolveResponse
        ] = await self.mapper_service.resolve(resolve_request)
        return (
            SyncResponseHelper.get_component().construct_success_sync_resolve_response(
                resolve_request,
                single_resolve_responses,
            )
        )

    async def unlink_sync(
        self,
        unlink_request: UnlinkRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
            RequestValidation.get_component().validate_request(unlink_request)
            RequestValidation.get_component().validate_unlink_request_header(
                unlink_request
            )
        except RequestValidationException as e:
            error_response = (
                SyncResponseHelper.get_component().construct_error_sync_response(
                    unlink_request, e
                )
            )
            return error_response

        single_unlink_responses: list[
            SingleResolveResponse
        ] = await self.mapper_service.unlink(unlink_request)
        return (
            SyncResponseHelper.get_component().construct_success_sync_unlink_response(
                unlink_request,
                single_unlink_responses,
            )
        )



import asyncio
import logging
import uuid
from typing import Annotated

import httpx
from fastapi import Depends
from openg2p_fastapi_common.controller import BaseController
from openg2p_g2pconnect_common_lib.jwt_signature_validator import JWTSignatureValidator
from openg2p_g2pconnect_common_lib.schemas import (
    AsyncCallbackRequest,
    AsyncResponse,
    Request,
)
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    ResolveRequest,
    SingleLinkResponse,
    SingleResolveResponse,
    SingleUnlinkResponse,
    SingleUpdateResponse,
    UnlinkRequest,
    UpdateRequest,
)

from ..config import Settings
from ..services import (
    AsyncResponseHelper,
    MapperService,
    RequestValidation,
    RequestValidationException,
)

_config = Settings.get_config()
_logger = logging.getLogger(_config.logging_default_logger_name)


async def _callback(
    async_call_back_request: AsyncCallbackRequest, url, url_suffix=None
):
    try:
        res = httpx.post(
            f"{url.rstrip('/')}{url_suffix}",
            headers={"content-type": "application/json"},
            content=async_call_back_request.model_dump_json(),
            timeout=_config.default_callback_timeout,
        )

        res.raise_for_status()
    except Exception as e:
        _logger.error(f"Error during callback: {e}")


class AsyncMapperController(BaseController):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mapper_service = MapperService.get_component()

        self.router.prefix += "/async"
        self.router.tags += ["G2PConnect Mapper Async"]

        self.action_to_method = {
            "link": self.mapper_service.link,
            "update": self.mapper_service.update,
            "resolve": self.mapper_service.resolve,
            "unlink": self.mapper_service.unlink,
        }

        self.router.add_api_route(
            "/link",
            self.link_async,
            responses={200: {"model": AsyncResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/update",
            self.update_async,
            responses={200: {"model": AsyncResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/resolve",
            self.resolve_async,
            responses={200: {"model": AsyncResponse}},
            methods=["POST"],
        )
        self.router.add_api_route(
            "/unlink",
            self.unlink_async,
            responses={200: {"model": AsyncResponse}},
            methods=["POST"],
        )

    async def link_async(
        self,
        link_request: LinkRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        correlation_id = str(uuid.uuid4())
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
        except RequestValidationException as e:
            error_response = (
                AsyncResponseHelper.get_component().construct_error_async_response(
                    link_request, e
                )
            )
            return error_response
        await asyncio.create_task(
            self.handle_service_and_link_callback(link_request, correlation_id, "link")
        )
        return AsyncResponseHelper.get_component().construct_success_async_response(
            link_request,
            correlation_id,
        )

    async def update_async(
        self,
        update_request: UpdateRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        correlation_id = str(uuid.uuid4())
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
        except RequestValidationException as e:
            error_response = (
                AsyncResponseHelper.get_component().construct_error_async_response(
                    update_request, e
                )
            )
            return error_response
        await asyncio.create_task(
            self.handle_service_and_update_callback(
                update_request, correlation_id, "update"
            )
        )
        return AsyncResponseHelper.get_component().construct_success_async_response(
            update_request,
            correlation_id,
        )

    async def resolve_async(
        self,
        resolve_request: ResolveRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        correlation_id = str(uuid.uuid4())
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
        except RequestValidationException as e:
            error_response = (
                AsyncResponseHelper.get_component().construct_error_async_response(
                    resolve_request, e
                )
            )
            return error_response
        await asyncio.create_task(
            self.handle_service_and_resolve_callback(
                resolve_request, correlation_id, "resolve"
            )
        )
        return AsyncResponseHelper.get_component().construct_success_async_response(
            resolve_request,
            correlation_id,
        )

    async def unlink_async(
        self,
        unlink_request: UnlinkRequest,
        is_signature_valid: Annotated[bool, Depends(JWTSignatureValidator())],
    ):
        correlation_id = str(uuid.uuid4())
        try:
            RequestValidation.get_component().validate_signature(is_signature_valid)
        except RequestValidationException as e:
            error_response = (
                AsyncResponseHelper.get_component().construct_error_async_response(
                    unlink_request, e
                )
            )
            return error_response
        try:
            RequestValidation.get_component().validate_request(unlink_request)
            RequestValidation.get_component().validate_unlink_async_request_header(
                unlink_request
            )
        except RequestValidationException as e:
            error_response = (
                AsyncResponseHelper.get_component().construct_error_async_response(
                    unlink_request, e
                )
            )
            return error_response
        await asyncio.create_task(
            self.handle_service_and_resolve_callback(
                unlink_request, correlation_id, "unlink"
            )
        )
        return AsyncResponseHelper.get_component().construct_success_async_response(
            unlink_request,
            correlation_id,
        )

    async def handle_service_and_link_callback(
        self,
        link_request: LinkRequest,
        correlation_id: str,
        action: str,
    ):
        try:
            RequestValidation.get_component().validate_async_request(link_request)
            RequestValidation.get_component().validate_link_async_request_header(
                link_request
            )
            single_link_responses: list[
                SingleLinkResponse
            ] = await self.action_to_method[action](link_request)

            async_call_back_request: (
                AsyncCallbackRequest
            ) = AsyncResponseHelper.get_component().construct_success_async_callback_link_request(
                link_request, correlation_id, single_link_responses
            )
            await self.make_callback(
                async_call_back_request,
                url=link_request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )
        except RequestValidationException as e:
            _logger.error(f"Error in handle_service_and_callback: {e}")
            error_response = AsyncResponseHelper.get_component().construct_error_async_callback_request(
                link_request, e
            )
            await self.make_callback(
                error_response,
                url=link_request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )

    async def handle_service_and_update_callback(
        self, request: Request, correlation_id: str, action: str
    ):
        try:
            RequestValidation.get_component().validate_async_request(request)
            RequestValidation.get_component().validate_update_async_request_header(
                request
            )
            single_update_responses: list[
                SingleUpdateResponse
            ] = await self.action_to_method[action](request)
            async_call_back_request: (
                AsyncCallbackRequest
            ) = AsyncResponseHelper.get_component().construct_success_async_callback_update_request(
                request, correlation_id, single_update_responses
            )
            await self.make_callback(
                async_call_back_request,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )
        except RequestValidationException as e:
            _logger.error(f"Error in handle_service_and_callback: {e}")
            error_response = AsyncResponseHelper.get_component().construct_error_async_callback_request(
                request, e
            )
            await self.make_callback(
                error_response,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )

    async def handle_service_and_resolve_callback(
        self, request: Request, correlation_id: str, action: str
    ):
        try:
            RequestValidation.get_component().validate_async_request(request)
            RequestValidation.get_component().validate_resolve_async_request_header(
                request
            )
            single_resolve_responses: list[
                SingleResolveResponse
            ] = await self.action_to_method[action](request)
            async_call_back_request: (
                AsyncCallbackRequest
            ) = AsyncResponseHelper.get_component().construct_success_async_callback_resolve_request(
                request, correlation_id, single_resolve_responses
            )
            await self.make_callback(
                async_call_back_request,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )
        except RequestValidationException as e:
            _logger.error(f"Error in handle_service_and_callback: {e}")
            error_response = AsyncResponseHelper.get_component().construct_error_async_callback_request(
                request, e
            )
            await self.make_callback(
                error_response,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )

    async def handle_service_and_unlink_callback(
        self, request: Request, correlation_id: str, action: str
    ):
        try:
            RequestValidation.get_component().validate_async_request(request)
            RequestValidation.get_component().validate_unlink_async_request_header(
                request
            )
            single_unlink_responses: list[
                SingleUnlinkResponse
            ] = await self.action_to_method[action](request)
            async_call_back_request: (
                AsyncCallbackRequest
            ) = AsyncResponseHelper.get_component().construct_success_async_callback_unlink_request(
                request, correlation_id, single_unlink_responses
            )
            await self.make_callback(
                async_call_back_request,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )
        except RequestValidationException as e:
            _logger.error(f"Error in handle_service_and_callback: {e}")
            error_response = AsyncResponseHelper.get_component().construct_error_async_callback_request(
                request, e
            )
            await self.make_callback(
                error_response,
                url=request.header.sender_uri,
                url_suffix=f"/on-{action}",
            )

    @staticmethod
    async def make_callback(
        async_call_back_request: AsyncCallbackRequest, url=None, url_suffix=None
    ):
        if not (url or _config.default_callback_url):
            return
        elif not url:
            url = _config.default_callback_url

        asyncio.ensure_future(
            _callback(async_call_back_request, url=url, url_suffix=url_suffix)
        )



from openg2p_g2pconnect_common_lib.schemas import StatusEnum
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkStatusReasonCode,
    ResolveStatusReasonCode,
    UnlinkStatusReasonCode,
    UpdateStatusReasonCode,
)


class LinkValidationException(Exception):
    def __init__(self, message, status, validation_error_type: LinkStatusReasonCode):
        self.message = message
        super().__init__(self.message)
        self.status: StatusEnum = status
        self.validation_error_type: LinkStatusReasonCode = validation_error_type


class UpdateValidationException(Exception):
    def __init__(self, message, status, validation_error_type: UpdateStatusReasonCode):
        self.message = message
        super().__init__(self.message)
        self.status: StatusEnum = status
        self.validation_error_type: UpdateStatusReasonCode = validation_error_type


class ResolveValidationException(Exception):
    def __init__(self, message, status, validation_error_type: ResolveStatusReasonCode):
        self.message = message
        super().__init__(self.message)
        self.status: StatusEnum = status
        self.validation_error_type: ResolveStatusReasonCode = validation_error_type


class UnlinkValidationException(Exception):
    def __init__(self, message, status, validation_error_type: UnlinkStatusReasonCode):
        self.message = message
        super().__init__(self.message)
        self.status: StatusEnum = status
        self.validation_error_type: UnlinkStatusReasonCode = validation_error_type


class RequestValidationException(Exception):
    # TODO : Add code
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(self.message)



from openg2p_fastapi_common.service import BaseService
from openg2p_g2pconnect_common_lib.schemas import (
    AsyncResponseStatusReasonCodeEnum,
    SyncResponseStatusReasonCodeEnum,
)

from .exceptions import RequestValidationException


class RequestValidation(BaseService):
    def validate_signature(self, is_signature_valid) -> None:
        if not is_signature_valid:
            raise RequestValidationException(
                code=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid,
                message=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid,
            )

        return None

    def validate_link_request_header(self, request) -> None:
        if request.header.action != "link":
            raise RequestValidationException(
                code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_update_request_header(self, request) -> None:
        if request.header.action != "update":
            raise RequestValidationException(
                code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_resolve_request_header(self, request) -> None:
        if request.header.action != "resolve":
            raise RequestValidationException(
                code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_unlink_request_header(self, request) -> None:
        if request.header.action != "unlink":
            raise RequestValidationException(
                code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_request(self, request) -> None:
        # TODO: Validate the request
        return None

    def validate_link_async_request_header(self, request) -> None:
        if request.header.action != "link":
            raise RequestValidationException(
                code=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_update_async_request_header(self, request) -> None:
        if request.header.action != "update":
            raise RequestValidationException(
                code=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_resolve_async_request_header(self, request) -> None:
        if request.header.action != "resolve":
            raise RequestValidationException(
                code=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_unlink_async_request_header(self, request) -> None:
        if request.header.action != "unlink":
            raise RequestValidationException(
                code=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
                message=AsyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
            )
        return None

    def validate_async_request(self, request) -> None:
        # TODO: Validate the request
        return None



from datetime import datetime
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from openg2p_g2pconnect_common_lib.schemas import (
    RequestHeader,
    StatusEnum,
    SyncResponseHeader,
    SyncResponseStatusReasonCodeEnum,
)
from openg2p_g2pconnect_mapper_lib.schemas import (
    LinkRequest,
    LinkRequestMessage,
    LinkResponse,
    LinkResponseMessage,
    SingleLinkRequest,
    SingleLinkResponse,
)
from openg2p_g2pconnect_mapper_lib.schemas.resolve import (
    ResolveRequest,
    ResolveRequestMessage,
    ResolveResponse,
    ResolveResponseMessage,
    SingleResolveRequest,
    SingleResolveResponse,
)
from openg2p_g2pconnect_mapper_lib.schemas.unlink import (
    SingleUnlinkRequest,
    SingleUnlinkResponse,
    UnlinkRequest,
    UnlinkRequestMessage,
    UnlinkResponse,
    UnlinkResponseMessage,
)
from openg2p_g2pconnect_mapper_lib.schemas.update import (
    SingleUpdateRequest,
    SingleUpdateResponse,
    UpdateRequest,
    UpdateRequestMessage,
    UpdateResponse,
    UpdateResponseMessage,
)
from openg2p_spar_mapper_api.controllers.sync_mapper_controller import (
    SyncMapperController,
)
from openg2p_spar_mapper_api.services import (
    RequestValidation,
    RequestValidationException,
)


def mock_validate_signature(is_signature_valid):
    if not is_signature_valid:
        raise RequestValidationException(
            code=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid,
            message=SyncResponseStatusReasonCodeEnum.rjct_jwt_invalid,
        )


@pytest.fixture
def setup_link_controller():
    controller = SyncMapperController()
    controller.mapper_service = AsyncMock()
    request_validation_mock = MagicMock()
    request_validation_mock.validate_request = MagicMock(return_value=True)
    request_validation_mock.validate_link_request_header = MagicMock(return_value=True)
    request_validation_mock.validate_signature = MagicMock(
        side_effect=mock_validate_signature
    )
    mock_link_response = LinkResponse(
        header=SyncResponseHeader(
            message_id="test_message_id",
            message_ts=datetime.now().isoformat(),
            action="link",
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message="Success",
        ),
        message=LinkResponseMessage(
            transaction_id="trans_id",
            link_response=[
                SingleLinkResponse(
                    reference_id="test_ref",
                    timestamp=datetime.now(),
                    status=StatusEnum.succ,
                    additional_info=[{}],
                    fa="test_fa",
                    status_reason_code=None,
                    status_reason_message="Test message",
                    locale="en",
                )
            ],
        ),
    )
    response_helper_link_mock = MagicMock()

    response_helper_link_mock.construct_success_sync_link_response.return_value = (
        mock_link_response
    )

    mock_error_link_response = LinkResponse(
        header=SyncResponseHeader(
            message_id="error_message_id",
            message_ts=datetime.now().isoformat(),
            action="error_action",
            status=StatusEnum.rjct,
            status_reason_code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported.value,
            status_reason_message="Validation error",
        ),
        message=LinkResponseMessage(
            transaction_id="error_trans_id",
            link_response=[],
        ),
    )
    response_helper_link_mock.construct_error_sync_response.return_value = (
        mock_error_link_response
    )

    with patch(
        "openg2p_spar_mapper_api.services.RequestValidation.get_component",
        return_value=request_validation_mock,
    ), patch(
        "openg2p_spar_mapper_api.services.SyncResponseHelper.get_component",
        return_value=response_helper_link_mock,
    ):
        mock_link_request = LinkRequest(
            header=RequestHeader(
                message_id="test_message_id",
                message_ts=datetime.now().isoformat(),
                action="test_action",
                sender_id="test_sender",
                total_count=1,
            ),
            message=LinkRequestMessage(
                transaction_id="test_transaction_id",
                link_request=[
                    SingleLinkRequest(
                        reference_id="test_ref",
                        timestamp=str(datetime.now()),
                        id="test_id",
                        fa="test_fa",
                    )
                ],
            ),
        )
        yield controller, mock_link_request


@pytest.fixture
def setup_update_controller():
    controller = SyncMapperController()
    controller.mapper_service = AsyncMock()

    request_validation_mock = MagicMock()
    request_validation_mock.validate_request = MagicMock(return_value=True)
    request_validation_mock.validate_update_request_header = MagicMock(
        return_value=True
    )
    request_validation_mock.validate_signature = MagicMock(
        side_effect=mock_validate_signature
    )
    mock_update_response = UpdateResponse(
        header=SyncResponseHeader(
            message_id="test_message_id",
            message_ts=datetime.now().isoformat(),
            action="update",
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message="Success",
        ),
        message=UpdateResponseMessage(
            transaction_id="trans_id",
            update_response=[
                SingleUpdateResponse(
                    reference_id="test_ref",
                    timestamp=datetime.now(),
                    status=StatusEnum.succ,
                    additional_info=[{}],
                    fa="test_fa",
                    status_reason_code=None,
                    status_reason_message="Test message",
                    locale="en",
                )
            ],
        ),
    )

    response_helper_update_mock = MagicMock()

    response_helper_update_mock.construct_success_sync_update_response.return_value = (
        mock_update_response
    )

    mock_error_update_response = UpdateResponse(
        header=SyncResponseHeader(
            message_id="error_message_id",
            message_ts=datetime.now().isoformat(),
            action="error_action",
            status=StatusEnum.rjct,
            status_reason_code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported.value,
            status_reason_message="Validation error",
        ),
        message=UpdateResponseMessage(
            transaction_id="error_trans_id",
            update_response=[],
        ),
    )

    response_helper_update_mock.construct_error_sync_response.return_value = (
        mock_error_update_response
    )

    with patch(
        "openg2p_spar_mapper_api.services.RequestValidation.get_component",
        return_value=request_validation_mock,
    ), patch(
        "openg2p_spar_mapper_api.services.SyncResponseHelper.get_component",
        return_value=response_helper_update_mock,
    ):
        mock_update_request = UpdateRequest(
            header=RequestHeader(
                message_id="test_message_id",
                message_ts=datetime.now().isoformat(),
                action="test_action",
                sender_id="test_sender",
                total_count=1,
            ),
            message=UpdateRequestMessage(
                transaction_id="test_transaction_id",
                update_request=[
                    SingleUpdateRequest(
                        reference_id="test_ref",
                        timestamp=str(datetime.now()),
                        id="test_id",
                        fa="test_fa",
                    )
                ],
            ),
        )
        yield controller, mock_update_request


@pytest.fixture
def setup_resolve_controller():
    controller = SyncMapperController()
    controller.mapper_service = AsyncMock()

    request_validation_mock = MagicMock()
    request_validation_mock.validate_request = MagicMock(return_value=True)
    request_validation_mock.validate_resolve_request_header = MagicMock(
        return_value=True
    )
    request_validation_mock.validate_signature = MagicMock(
        side_effect=mock_validate_signature
    )
    mock_resolve_response = ResolveResponse(
        header=SyncResponseHeader(
            message_id="test_message_id",
            message_ts=datetime.now().isoformat(),
            action="resolve",
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message="Success",
        ),
        message=ResolveResponseMessage(
            transaction_id="trans_id",
            resolve_response=[
                SingleResolveResponse(
                    reference_id="test_ref",
                    timestamp=datetime.now(),
                    status=StatusEnum.succ,
                    additional_info=[{}],
                    fa="test_fa",
                    status_reason_code=None,
                    status_reason_message="Test message",
                    locale="en",
                )
            ],
        ),
    )

    response_helper_resolve_mock = MagicMock()

    response_helper_resolve_mock.construct_success_sync_resolve_response.return_value = (
        mock_resolve_response
    )

    mock_error_resolve_response = ResolveResponse(
        header=SyncResponseHeader(
            message_id="error_message_id",
            message_ts=datetime.now().isoformat(),
            action="error_action",
            status=StatusEnum.rjct,
            status_reason_code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported.value,
            status_reason_message="Validation error",
        ),
        message=ResolveResponseMessage(
            transaction_id="error_trans_id",
            resolve_response=[],
        ),
    )

    response_helper_resolve_mock.construct_error_sync_response.return_value = (
        mock_error_resolve_response
    )

    with patch(
        "openg2p_spar_mapper_api.services.RequestValidation.get_component",
        return_value=request_validation_mock,
    ), patch(
        "openg2p_spar_mapper_api.services.SyncResponseHelper.get_component",
        return_value=response_helper_resolve_mock,
    ):
        mock_resolve_request = ResolveRequest(
            header=RequestHeader(
                message_id="test_message_id",
                message_ts=datetime.now().isoformat(),
                action="test_action",
                sender_id="test_sender",
                total_count=1,
            ),
            message=ResolveRequestMessage(
                transaction_id="test_transaction_id",
                resolve_request=[
                    SingleResolveRequest(
                        reference_id="test_ref",
                        timestamp=str(datetime.now()),
                        id="test_id",
                        fa="test_fa",
                    )
                ],
            ),
        )
        yield controller, mock_resolve_request


@pytest.fixture
def setup_unlink_controller():
    controller = SyncMapperController()
    controller.mapper_service = AsyncMock()

    request_validation_mock = MagicMock()
    request_validation_mock.validate_request = MagicMock(return_value=True)
    request_validation_mock.validate_unlink_request_header = MagicMock(
        return_value=True
    )
    request_validation_mock.validate_signature = MagicMock(
        side_effect=mock_validate_signature
    )
    mock_unlink_response = UnlinkResponse(
        header=SyncResponseHeader(
            message_id="test_message_id",
            message_ts=datetime.now().isoformat(),
            action="unlink",
            status=StatusEnum.succ,
            status_reason_code=None,
            status_reason_message="Success",
        ),
        message=UnlinkResponseMessage(
            transaction_id="trans_id",
            unlink_response=[
                SingleUnlinkResponse(
                    reference_id="test_ref",
                    timestamp=datetime.now(),
                    status=StatusEnum.succ,
                    additional_info=[{}],
                    fa="test_fa",
                    status_reason_code=None,
                    status_reason_message="Test message",
                    locale="en",
                )
            ],
        ),
    )

    response_helper_unlink_mock = MagicMock()

    response_helper_unlink_mock.construct_success_sync_unlink_response.return_value = (
        mock_unlink_response
    )

    mock_error_unlink_response = UnlinkResponse(
        header=SyncResponseHeader(
            message_id="error_message_id",
            message_ts=datetime.now().isoformat(),
            action="error_action",
            status=StatusEnum.rjct,
            status_reason_code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported.value,
            status_reason_message="Validation error",
        ),
        message=UnlinkResponseMessage(
            transaction_id="error_trans_id",
            unlink_response=[],
        ),
    )

    response_helper_unlink_mock.construct_error_sync_response.return_value = (
        mock_error_unlink_response
    )

    with patch(
        "openg2p_spar_mapper_api.services.RequestValidation.get_component",
        return_value=request_validation_mock,
    ), patch(
        "openg2p_spar_mapper_api.services.SyncResponseHelper.get_component",
        return_value=response_helper_unlink_mock,
    ):
        mock_unlink_request = UnlinkRequest(
            header=RequestHeader(
                message_id="test_message_id",
                message_ts=datetime.now().isoformat(),
                action="test_action",
                sender_id="test_sender",
                total_count=1,
            ),
            message=UnlinkRequestMessage(
                transaction_id="test_transaction_id",
                unlink_request=[
                    SingleUnlinkRequest(
                        reference_id="test_ref",
                        timestamp=str(datetime.now()),
                        id="test_id",
                        fa="test_fa",
                    )
                ],
            ),
        )
        yield controller, mock_unlink_request


@pytest.mark.asyncio
async def test_link_sync_success(setup_link_controller):
    controller, mock_link_request = setup_link_controller
    assert controller is not None

    response = await controller.link_sync(mock_link_request, is_signature_valid=True)
    assert response.header.status == StatusEnum.succ
    assert response.message.transaction_id == "trans_id"
    controller.mapper_service.link.assert_called_once_with(mock_link_request)


@pytest.mark.asyncio
async def test_link_sync_invalid_signature(setup_link_controller):
    controller, mock_link_request = setup_link_controller
    assert controller is not None

    response = await controller.link_sync(mock_link_request, is_signature_valid=False)
    assert response.header.status == StatusEnum.rjct
    assert response.header.status_reason_message == "Validation error"


@pytest.mark.asyncio
async def test_update_sync_success(setup_update_controller):
    controller, mock_update_request = setup_update_controller
    assert controller is not None
    response = await controller.update_sync(
        mock_update_request, is_signature_valid=True
    )
    assert response.header.status == StatusEnum.succ
    assert response.message.transaction_id == "trans_id"
    controller.mapper_service.update.assert_called_once_with(mock_update_request)


@pytest.mark.asyncio
async def test_update_sync_invalid_signature(setup_update_controller):
    controller, mock_update_request = setup_update_controller
    assert controller is not None

    response = await controller.update_sync(
        mock_update_request, is_signature_valid=False
    )
    assert response.header.status == StatusEnum.rjct
    assert response.header.status_reason_message == "Validation error"


@pytest.mark.asyncio
async def test_resolve_sync_success(setup_resolve_controller):
    controller, mock_resolve_request = setup_resolve_controller
    assert controller is not None
    response = await controller.resolve_sync(
        mock_resolve_request, is_signature_valid=True
    )
    assert response.header.status == StatusEnum.succ
    assert response.message.transaction_id == "trans_id"
    controller.mapper_service.resolve.assert_called_once_with(mock_resolve_request)


@pytest.mark.asyncio
async def test_resolve_sync_invalid_signature(setup_resolve_controller):
    controller, mock_resolve_request = setup_resolve_controller
    assert controller is not None

    response = await controller.resolve_sync(
        mock_resolve_request, is_signature_valid=False
    )
    assert response.header.status == StatusEnum.rjct
    assert response.header.status_reason_message == "Validation error"


@pytest.mark.asyncio
async def test_unlink_sync_success(setup_unlink_controller):
    controller, mock_unlink_request = setup_unlink_controller
    assert controller is not None
    response = await controller.unlink_sync(
        mock_unlink_request, is_signature_valid=True
    )
    assert response.header.status == StatusEnum.succ
    assert response.message.transaction_id == "trans_id"
    controller.mapper_service.unlink.assert_called_once_with(mock_unlink_request)


@pytest.mark.asyncio
async def test_unlink_sync_invalid_signature(setup_unlink_controller):
    controller, mock_unlink_request = setup_unlink_controller
    assert controller is not None

    response = await controller.unlink_sync(
        mock_unlink_request, is_signature_valid=False
    )
    assert response.header.status == StatusEnum.rjct
    assert response.header.status_reason_message == "Validation error"


@pytest.mark.asyncio
async def test_link_sync_validation_error(setup_link_controller):
    controller, mock_link_request = setup_link_controller
    validation_error = RequestValidationException(
        code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
        message="Validation error",
    )
    with patch.object(
        RequestValidation.get_component(),
        "validate_request",
        side_effect=validation_error,
    ), patch.object(
        RequestValidation.get_component(),
        "validate_link_request_header",
        side_effect=validation_error,
    ):
        response = await controller.link_sync(
            mock_link_request, is_signature_valid=True
        )
        assert response.header.status == StatusEnum.rjct
        assert validation_error.message in response.header.status_reason_message
        controller.mapper_service.link.assert_not_called()


@pytest.mark.asyncio
async def test_update_sync_validation_error(setup_update_controller):
    controller, mock_update_request = setup_update_controller
    validation_error = RequestValidationException(
        code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
        message="Validation error",
    )
    with patch.object(
        RequestValidation.get_component(),
        "validate_request",
        side_effect=validation_error,
    ), patch.object(
        RequestValidation.get_component(),
        "validate_update_request_header",
        side_effect=validation_error,
    ):
        response = await controller.update_sync(
            mock_update_request, is_signature_valid=True
        )
        assert response.header.status == StatusEnum.rjct
        assert validation_error.message in response.header.status_reason_message
        controller.mapper_service.update.assert_not_called()


@pytest.mark.asyncio
async def test_resolve_sync_validation_error(setup_resolve_controller):
    controller, mock_resolve_request = setup_resolve_controller
    validation_error = RequestValidationException(
        code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
        message="Validation error",
    )
    with patch.object(
        RequestValidation.get_component(),
        "validate_request",
        side_effect=validation_error,
    ), patch.object(
        RequestValidation.get_component(),
        "validate_resolve_request_header",
        side_effect=validation_error,
    ):
        response = await controller.resolve_sync(
            mock_resolve_request, is_signature_valid=True
        )
        assert response.header.status == StatusEnum.rjct
        assert validation_error.message in response.header.status_reason_message
        controller.mapper_service.resolve.assert_not_called()


@pytest.mark.asyncio
async def test_unlink_sync_validation_error(setup_unlink_controller):
    controller, mock_unlink_request = setup_unlink_controller
    validation_error = RequestValidationException(
        code=SyncResponseStatusReasonCodeEnum.rjct_action_not_supported,
        message="Validation error",
    )
    with patch.object(
        RequestValidation.get_component(),
        "validate_request",
        side_effect=validation_error,
    ), patch.object(
        RequestValidation.get_component(),
        "validate_unlink_request_header",
        side_effect=validation_error,
    ):
        response = await controller.unlink_sync(
            mock_unlink_request, is_signature_valid=True
        )
        assert response.header.status == StatusEnum.rjct
        assert validation_error.message in response.header.status_reason_message
        controller.mapper_service.unlink.assert_not_called()



from .exceptions import LinkValidationException, RequestValidationException
from .id_fa_mapping_validations import IdFaMappingValidations
from .mapper import MapperService
from .request_helper import AsyncRequestHelper, SyncRequestHelper
from .request_validations import RequestValidation
from .response_helper import AsyncResponseHelper, SyncResponseHelper
from .session_service import SessionInitializer
